from typing import Any
import urllib.parse
from collections import defaultdict
import json

from django.contrib import admin
from django.forms import ModelForm
from django import forms
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.contrib.admin.views.main import ChangeList
from django.urls import path
from django.http import JsonResponse, HttpResponse
from django.contrib.admin.utils import unquote
from django.core.exceptions import PermissionDenied
from django.utils.text import capfirst
from django.template.response import TemplateResponse
from django.contrib.messages import add_message, ERROR
from django.utils.translation import gettext_lazy as _
from django.db.models import Q
from django.http import HttpResponseRedirect
from django.contrib import messages
from django.shortcuts import resolve_url
from django.shortcuts import get_object_or_404
from django.utils.html import escape
from django.core.exceptions import ValidationError

from .models import (
    ContentSetting,
    HistoryContentSetting,
    UserTagSetting,
    UserPreview,
    UserPreviewHistory,
)
from .context_managers import content_settings_context
from .conf import (
    USER_DEFINED_TYPES_INSTANCE,
    USER_DEFINED_TYPES_INITIAL,
    content_settings,
    validate_all_with_context,
)
from .settings import (
    USER_TAGS,
    USER_DEFINED_TYPES,
    PREVIEW_ON_SITE_HREF,
    ADMIN_CHECKSUM_CHECK_BEFORE_SAVE,
    UI_DOC_URL,
    PREVIEW_ON_SITE_SHOW,
)
from .caching import get_type_by_name
from .utils import class_names
from .export import export_to_format, preview_data, import_to


def user_able_to_update(user, name, user_defined_type=None):
    cs_type = get_type_by_name(name)
    if cs_type is None and user_defined_type is not None:
        cs_type = USER_DEFINED_TYPES_INSTANCE[user_defined_type]()
    if cs_type is None:
        return False
    return cs_type.can_update(user)


TAGS_PARAM = "tags"
TAGS_SPLITER = "|"


def get_selected_tags_from_params(params):
    if TAGS_PARAM not in params:
        return set()
    return set(params[TAGS_PARAM].split(TAGS_SPLITER))


def html_classes(name):
    classes = "</li><li>".join(
        f"{cls_name} <i>from {module_name}</i>"
        for cls_name, module_name in class_names(get_type_by_name(name).__class__)
    )
    if not classes:
        return ""

    return mark_safe(
        f"<div><a class='replace_with_ul'>...</a> <ul style='display: none; margin-top: 1.5em'><li>{classes}</li></ul></div>"
    )


class SettingsChangeList(ChangeList):
    def get_filters_params(self, *args, **kwargs):
        params = super().get_filters_params(*args, **kwargs)
        if TAGS_PARAM in params:
            del params[TAGS_PARAM]
        return params

    def get_queryset(self, request, *args, **kwargs):
        from .conf import get_type_by_name

        q = super().get_queryset(request, *args, **kwargs)
        user = request.user

        tags = get_selected_tags_from_params(self.params)
        if tags:

            user_settings = list(
                UserTagSetting.objects.filter(
                    user=request.user, tag__in=tags
                ).values_list("name", flat=True)
            )
            if user_settings:
                q = q.filter(name__in=user_settings)

            combine = Q()
            for tag in tags:
                if tag in USER_TAGS:
                    continue
                combine &= Q(tags__iregex=rf"(^|\n){tag}($|\n)")

            q = q.filter(combine)
        # This is terrable
        # But I don't know how to do it better
        # I need to filter out the names user don't have permissions to see

        names = list(q.values_list("name", flat=True))
        return q.filter(
            ~Q(
                name__in=[
                    name
                    for name in names
                    if not get_type_by_name(name)
                    or get_type_by_name(name).constant
                    or not get_type_by_name(name).can_view(user)
                ]
            )
        )


class ContentSettingForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(ContentSettingForm, self).__init__(*args, **kwargs)
        if self.instance.name:
            cs_type = get_type_by_name(self.instance.name)
            if cs_type:
                self.fields["value"] = cs_type.get_field()
        self.fields["value"].strip = False

        if "user_defined_type" in self.fields:
            self.fields["user_defined_type"].widget = forms.Select(
                choices=[(v[0], v[2]) for v in USER_DEFINED_TYPES]
            )

    class Meta:
        model = ContentSetting
        fields = ["value"]


# we want to have a separate form for a single edit interface
# because we want to validate the chain of content setting changes
class ContentSettingFormWithChainValidation(ContentSettingForm):
    def clean(self):
        ret = super().clean()
        # import ipdb; ipdb.set_trace()
        if self.instance and not self.errors:
            validate_all_with_context(
                {
                    (self.cleaned_data.get("name") or self.instance.name): (
                        (
                            self.cleaned_data["value"],
                            self.cleaned_data.get("user_defined_type"),
                            self.cleaned_data.get("tags"),
                            self.cleaned_data.get("help"),
                        )
                        if "user_defined_type" in self.cleaned_data
                        else self.cleaned_data["value"]
                    )
                }
            )
        return ret


class ContentSettingAdmin(admin.ModelAdmin):
    list_display = ["name", "value", "setting_help_and_tags", "marks"]
    list_editable = [
        "value",
    ]
    search_fields = ["name", "value", "tags", "help"]
    readonly_fields = [
        "py_data",
        "name",
        "setting_help",
        "setting_tags",
        "version",
        "default_value",
    ]
    history_list_display = ["value"]
    form = ContentSettingFormWithChainValidation
    actions = ["export_as_json", "view_as_json"]

    def export_as_json(self, request, queryset, download=True):
        json_data = export_to_format(
            row for row in queryset if get_type_by_name(row.name).can_view(request.user)
        )
        response = HttpResponse(content_type="application/json")
        if download:
            response["Content-Disposition"] = (
                'attachment; filename="content_settings.json"'
            )
        response.write(json_data)
        return response

    export_as_json.short_description = _("Export selected content settings")

    def view_as_json(self, request, queryset):
        return self.export_as_json(request, queryset, download=False)

    view_as_json.short_description = _("View selected content settings as JSON")

    def save_model(self, request, obj, form, change):
        if user_able_to_update(request.user, obj.name, obj.user_defined_type):
            prev_obj = ContentSetting.objects.filter(pk=obj.pk).first()
            if prev_obj and not any(
                getattr(prev_obj, name) != getattr(obj, name)
                for name in ("value", "tags", "help", "user_defined_type", "name")
            ):
                return

            if "_preview_on_site" in request.POST:
                if obj.user_defined_type:
                    UserPreview.add_by_user(
                        user=request.user,
                        name=obj.name,
                        value=obj.value,
                        user_defined_type=obj.user_defined_type,
                        tags=obj.tags_set,
                        help=obj.help,
                    )
                else:
                    UserPreview.add_by_user(
                        user=request.user,
                        name=obj.name,
                        value=obj.value,
                    )
            else:
                super().save_model(request, obj, form, change)
                HistoryContentSetting.update_last_record_for_name(
                    obj.name, request.user
                )
        else:
            add_message(
                request,
                ERROR,
                _("You are not allowed to change %(name)s") % {"name": obj.name},
            )

    def get_readonly_fields(self, request, obj=None):
        if not USER_DEFINED_TYPES or obj and not obj.user_defined_type:
            return super().get_readonly_fields(request, obj)

        return [
            v
            for v in super().get_readonly_fields(request, obj)
            if v
            not in (
                "name",
                "setting_help",
                "setting_tags",
            )
        ]

    def get_form(self, request, obj=None, change=False, **kwargs):
        if USER_DEFINED_TYPES and not obj or obj.user_defined_type:
            kwargs["fields"] = ["user_defined_type", "name", "value", "help", "tags"]
        else:
            kwargs["fields"] = [
                "value",
            ]

        return super().get_form(request, obj=None, change=False, **kwargs)

    def get_changelist(self, request, **kwargs):
        return SettingsChangeList

    def context_preview_settings(self, request):
        if not PREVIEW_ON_SITE_SHOW or not request.user.has_perm(
            "content_settings.can_preview_on_site"
        ):
            return {}

        return {
            "preview_settings": UserPreview.objects.filter(user=request.user),
            "PREVIEW_ON_SITE_HREF": PREVIEW_ON_SITE_HREF,
            "PREVIEW_ON_SITE_SHOW": True,
        }

    def context_tags(self, request):
        from .conf import get_type_by_name

        extra_context = {}
        selected_tags = get_selected_tags_from_params(request.GET)

        if "q" in request.GET:
            init_q = "?q=" + urllib.parse.quote(request.GET["q"]) + "&"
        else:
            init_q = "?"

        def q_tags(tags):
            if not tags:
                return init_q
            return (
                init_q + TAGS_PARAM + "=" + urllib.parse.quote(TAGS_SPLITER.join(tags))
            )

        def q_add_tag(tag):
            return q_tags(selected_tags | set([tag]))

        def q_remove_tag(tag):
            return q_tags(selected_tags - set([tag]))

        extra_context["selected_tags"] = [
            {
                "tag": USER_TAGS[tag][0] if tag in USER_TAGS else tag,
                "url": q_remove_tag(tag),
            }
            for tag in sorted(selected_tags)
        ]

        tags_stat = defaultdict(int)
        user_settings = UserTagSetting.get_user_settings(request.user)

        for cs in self.get_changelist_instance(request).get_queryset(request):
            cs_type = get_type_by_name(cs.name)
            if not cs_type or not cs_type.can_view(request.user) or cs_type.constant:
                continue
            val_tags = cs.tags_set | user_settings[cs.name]
            if selected_tags and (not val_tags or selected_tags - val_tags):
                continue

            for tag in val_tags - selected_tags:
                tags_stat[tag] += 1

        extra_context["available_tags"] = [
            {
                "tag": USER_TAGS[tag][0] if tag in USER_TAGS else tag,
                "url": q_add_tag(tag),
                "total": total,
            }
            for tag, total in sorted(
                tags_stat.items(), key=lambda a: a[0] and a[0] not in USER_TAGS
            )
        ]

        return extra_context

    def form_checksum_valid(self, request):
        if (
            ADMIN_CHECKSUM_CHECK_BEFORE_SAVE
            and request.POST["content_settings_form_checksum"]
            != content_settings.form_checksum
        ):
            add_message(
                request,
                ERROR,
                _(
                    "The content settings have been changed by another user. Please reload the page and try again. (or disable ADMIN_CHECKSUM_CHECK_BEFORE_SAVE in settings)"
                ),
            )
            return False
        return True

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if request.method == "POST" and not self.form_checksum_valid(request):
            return HttpResponseRedirect(request.path)
        return super().changeform_view(
            request,
            object_id=object_id,
            form_url=form_url,
            extra_context={
                **(extra_context or {}),
                **self.context_preview_settings(request),
            },
        )

    def get_changelist_formset(self, request, **kwargs):
        # validation of changing this particular value didn't change other values in the chain
        # this is done by using validate_all_with_context
        BaseFormset = super().get_changelist_formset(request, **kwargs)

        class ContentSettingFormset(BaseFormset):
            def clean(self):
                ret = super().clean()
                if hasattr(self, "cleaned_data"):
                    validate_all_with_context(
                        {data["id"].name: data["value"] for data in self.cleaned_data}
                    )
                return ret

        return ContentSettingFormset

    def changelist_view(self, request, extra_context=None):
        if request.method == "POST" and not self.form_checksum_valid(request):
            return HttpResponseRedirect(request.path)

        def key_value_from_request():
            total_forms = int(request.POST["form-TOTAL_FORMS"])

            for i in range(total_forms):
                yield ContentSetting.objects.get(
                    id=request.POST[f"form-{i}-id"]
                ).name, request.POST[f"form-{i}-value"]

        super_changelist = super().changelist_view

        def gen_response():
            return super_changelist(
                request,
                {
                    **(extra_context or {}),
                    **self.context_tags(request),
                    **self.context_preview_settings(request),
                    "UI_DOC_URL": UI_DOC_URL,
                },
            )

        if not request.method == "POST" or "_save" not in request.POST:
            return gen_response()

        with content_settings_context(
            _raise_errors=False,
            **{name: value for name, value in key_value_from_request()},
        ):
            return gen_response()

    def context_tags_view(self, request):
        return TemplateResponse(
            request,
            "admin/content_settings/contentsetting/context_tags.html",
            self.context_tags(request),
        )

    def change_view(self, request, object_id, form_url="", extra_context=None):
        from .conf import get_type_by_name

        cs = ContentSetting.objects.filter(pk=object_id).first()

        cs_type = get_type_by_name(cs.name)
        if cs and not cs_type or not cs_type.can_view(request.user) or cs_type.constant:
            raise PermissionDenied

        extra_context = {
            "PREVIEW_ON_SITE_SHOW": PREVIEW_ON_SITE_SHOW
            and request.user.has_perm("content_settings.can_preview_on_site"),
            **(extra_context or {}),
        }

        return super().change_view(request, object_id, form_url, extra_context)

    def history_view(self, request, object_id, extra_context=None):
        from .conf import get_type_by_name

        # First check if the user can see this history.
        model = self.model
        obj = self.get_object(request, unquote(object_id))
        if obj is None:
            return self._get_obj_does_not_exist_redirect(
                request, model._meta, object_id
            )

        cs_type = get_type_by_name(obj.name)
        if (
            not self.has_view_or_change_permission(request, obj)
            or not cs_type
            or not cs_type.can_view_history(request.user)
            or cs_type.constant
        ):
            raise PermissionDenied

        opts = model._meta
        app_label = opts.app_label

        context = {
            **self.admin_site.each_context(request),
            "title": f"Change history: {obj}",
            "subtitle": None,
            "module_name": str(capfirst(opts.verbose_name_plural)),
            "object": obj,
            "opts": opts,
            "preserved_filters": self.get_preserved_filters(request),
            "action_list": HistoryContentSetting.gen_unique_records(obj.name),
            **(extra_context or {}),
        }

        request.current_app = self.admin_site.name

        return TemplateResponse(
            request, f"admin/{app_label}/{opts.model_name}/object_history.html", context
        )

    def get_changelist_form(self, request, **kwargs):
        return ContentSettingForm

    def has_delete_permission(self, request, obj=None):
        default_del_permission = super().has_delete_permission(request, obj)
        if not default_del_permission or not USER_DEFINED_TYPES:
            return False
        return not obj or obj.user_defined_type is not None

    def has_add_permission(self, *args):
        return super().has_add_permission(*args) and USER_DEFINED_TYPES

    def has_change_permission(self, request, obj=None):
        default_change_permission = super().has_change_permission(request, obj)
        if not default_change_permission:
            return False
        if not obj:
            return True
        return user_able_to_update(request.user, obj.name)

    def setting_help(self, obj):
        cs_type = get_type_by_name(obj.name)
        if cs_type is None or cs_type.constant:
            return _("(the setting is not using)")

        if obj.user_defined_type:
            help = cs_type.get_help()
        else:
            help = obj.help
        return mark_safe(help + html_classes(obj.name))

    setting_help.short_description = _("Help")

    def setting_help_and_tags(self, obj):
        return mark_safe(
            f"<div style='min-width: 300px;'>{self.setting_help(obj)}<br>{self.setting_tags(obj)}</div>"
        )

    setting_help_and_tags.short_description = _("Help")

    def setting_tags(self, obj):
        cs_type = get_type_by_name(obj.name)
        if cs_type is None or cs_type.constant:
            return _("(the setting is not using)")

        tags = obj.tags_set

        if not tags:
            return "---"
        return ", ".join(tags)

    setting_tags.short_description = _("Tags")

    def marks(self, obj):
        return mark_safe(
            "".join(
                f"""<span>
            <span class="user_tag user_tag_selected user_tag_hidden" data-tag="{tag}" data-name="{obj.name}">{marks[0]}</span>
            <span class="user_tag user_tag_hidden" data-tag="{tag}" data-name="{obj.name}">{marks[1]}</span>
            </span>"""
                for tag, marks in USER_TAGS.items()
            )
        )

    def py_data(self, obj):
        cs_type = get_type_by_name(obj.name)

        if cs_type is None or cs_type.constant:
            return ""

        full_value = cs_type.get_full_admin_preview_value(obj.value, obj.name)
        if "error" in full_value:
            return mark_safe(f"<pre style='color: red'>{full_value['error']}</pre>")
        if "html" in full_value:
            return mark_safe(full_value["html"])

    def default_value(self, obj):
        cs_type = get_type_by_name(obj.name)

        if cs_type is None or cs_type.constant:
            return ""

        attr_default = (
            cs_type.default.replace("&", "&amp;")
            .replace("\n", "&NewLine;")
            .replace('"', "&quot;")
        )

        label = _('Reset ("save" is required for applying)')
        return mark_safe(
            f"""
        <pre class="default_value">{escape(cs_type.default)}</pre>
        <a class="reset_default" data-value="{attr_default}" style="cursor: pointer; color:var(--link-fg)">{label}</a>
        """
        )

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "preview/",
                self.admin_site.admin_view(self.preview_setting),
                name="content_settings_preview_setting",
            ),
            path(
                "add-tag/",
                self.admin_site.admin_view(self.add_tag),
                name="content_settings_add_tag",
            ),
            path(
                "remove-tag/",
                self.admin_site.admin_view(self.remove_tag),
                name="content_settings_remove_tag",
            ),
            path(
                "context-tags/",
                self.admin_site.admin_view(self.context_tags_view),
                name="content_settings_context_tags",
            ),
            path(
                "apply-preview-on-site/",
                self.admin_site.admin_view(self.apply_preview_on_site),
                name="content_settings_apply_preview_on_site",
            ),
            path(
                "reset-preview-on-site/",
                self.admin_site.admin_view(self.reset_preview_on_site),
                name="content_settings_reset_preview_on_site",
            ),
            path(
                "remove-preview-on-site/",
                self.admin_site.admin_view(self.remove_preview_on_site),
                name="content_settings_remove_preview_on_site",
            ),
            path(
                "import-json/",
                self.admin_site.admin_view(self.import_json_view),
                name="content_settings_import_json",
            ),
        ]
        return custom_urls + urls

    def remove_preview_on_site(self, request):
        preview_setting = get_object_or_404(
            UserPreview, user=request.user, name=request.GET["name"]
        )
        UserPreviewHistory.user_record(
            preview_setting, UserPreviewHistory.STATUS_REMOVED
        )
        preview_setting.delete()
        self.message_user(request, _("Preview setting removed"), messages.SUCCESS)
        return HttpResponseRedirect(
            request.META.get("HTTP_REFERER", resolve_url("admin:index"))
        )

    def apply_preview_on_site(self, request):
        """
        The view that applies collected preview settings to the actual settings.

        Consist of the following stages:

        * validate each value independently
        * previous value was updated before applying
        * validate the whole chain of values
        * apply the values

        Edge cases:

        * preview settings will be automatically removed in case the value is not longer passible to apply
        * if setting was changed and now it is equal to preview setting - the preview will be removed
        """

        def redirect_back():
            return HttpResponseRedirect(
                request.META.get(
                    "HTTP_REFERER",
                    resolve_url("admin:content_settings_contentsetting_changelist"),
                )
            )

        previews = list(UserPreview.objects.filter(user=request.user))
        preview_context = {
            p.name: (
                p.value
                if not p.user_defined_type
                else (p.value, p.user_defined_type, p.tags, p.help)
            )
            for p in previews
        }

        if not preview_context:
            self.message_user(
                request, _("No preview settings to apply"), messages.WARNING
            )
            return redirect_back()

        apply = True
        for name, value in preview_context.items():
            preview = UserPreview.objects.get(user=request.user, name=name)
            is_userdefined = isinstance(value, tuple)

            setting = ContentSetting.objects.filter(name=name).first()
            if not setting and not is_userdefined:
                UserPreviewHistory.user_record(
                    preview, UserPreviewHistory.STATUS_IGNORED
                )
                preview.delete()
                self.message_user(
                    request,
                    _(
                        "Preview setting %s removed because the setting itself doesn't exist"
                    )
                    % name,
                    messages.WARNING,
                )
                apply = False
                continue

            cs_type = (
                USER_DEFINED_TYPES_INITIAL.get(value[1])
                if is_userdefined
                else get_type_by_name(name)
            )
            if not cs_type:
                UserPreviewHistory.user_record(
                    preview, UserPreviewHistory.STATUS_IGNORED
                )
                preview.delete()
                self.message_user(
                    request,
                    _("Setting Type %s doesn't exist") % name,
                    messages.ERROR,
                )
                apply = False
                continue

            if not cs_type.can_update(request.user):
                self.message_user(
                    request,
                    _("You don't have permissions to update the setting %s") % name,
                    messages.ERROR,
                )
                apply = False
                continue

            try:
                cs_type.validate_value(value[0] if is_userdefined else value)
            except ValidationError as e:
                self.message_user(request, name + ": " + str(e), messages.ERROR)
                apply = False
                continue

            if not is_userdefined and preview.from_value != setting.value:
                if setting.value == preview.value:
                    UserPreviewHistory.user_record(
                        preview, UserPreviewHistory.STATUS_IGNORED
                    )
                    preview.delete()
                    self.message_user(
                        request,
                        _("The value %s was changed already applied") % name,
                        messages.WARNING,
                    )
                else:
                    self.message_user(
                        request,
                        _(
                            "The value %s was changed before applying the preview (check and apply again)"
                        )
                        % name,
                        messages.WARNING,
                    )
                    preview.from_value = setting.value
                    preview.save()
                apply = False
                continue

        if not apply:
            return redirect_back()
        try:
            validate_all_with_context(preview_context)
        except ValidationError as e:
            self.message_user(request, str(e), messages.ERROR)
            return redirect_back()

        for preview in previews:
            preview.apply()
            preview.delete()
        self.message_user(request, _("Preview settings applied"), messages.SUCCESS)

        return redirect_back()

    def reset_preview_on_site(self, request):
        for preview_setting in UserPreview.objects.filter(user=request.user):
            UserPreviewHistory.user_record(
                preview_setting, UserPreviewHistory.STATUS_REMOVED
            )
            preview_setting.delete()

        self.message_user(request, _("Preview settings removed"), messages.SUCCESS)
        return HttpResponseRedirect(
            request.META.get("HTTP_REFERER", resolve_url("admin:index"))
        )

    # ajax view that creates UserTagSetting
    def add_tag(self, request):
        try:
            name = request.POST["name"]
            tag = request.POST["tag"]
            UserTagSetting.objects.create(
                user=request.user,
                name=name,
                tag=tag,
            )
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        else:
            return JsonResponse(
                {
                    "success": True,
                }
            )

    # ajax view that removes UserTagSetting
    def remove_tag(self, request):
        try:
            name = request.POST["name"]
            tag = request.POST["tag"]
            UserTagSetting.objects.get(
                user=request.user,
                name=name,
                tag=tag,
            ).delete()
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        else:
            return JsonResponse(
                {
                    "success": True,
                }
            )

    def import_json_view(self, request):
        """
        Importing Content Settings from different sources using JSON format.

        The sources of the format can be different:
        * JSON file
        * JSON text field
        * ids of HistoryContentSetting records

        The view assumes that all of the data is migrated and values in database are aligned with the types.

        The view has too many lines of code for several reasons:
        * the code includes a lot of check for the json format itself
        * the view inlcudes two actions - peview results and apply results
        * the view inludes logic for user defined types
        """
        raw_json = '{"settings": {}}'
        if request.method == "POST":
            if "json_file" in request.FILES:
                raw_json = request.FILES["json_file"].read().decode("utf-8")
            elif "raw_json" in request.POST:
                raw_json = request.POST["raw_json"]
        else:
            if "history_ids" in request.GET:
                history_ids = request.GET["history_ids"].split(",")
                history_records = HistoryContentSetting.objects.filter(
                    id__in=[int(id) for id in history_ids]
                )
                raw_json = export_to_format(history_records)

        core_context = {
            "title": _("Import Content Settings"),
            "opts": self.model._meta,
        }

        def response_error(message):
            self.message_user(request, message, messages.ERROR)
            return TemplateResponse(
                request,
                "admin/content_settings/contentsetting/import_json.html",
                {**core_context, "raw_json": raw_json},
            )

        try:
            data = json.loads(raw_json)
        except Exception as e:
            return response_error(_("Error JSON parsing: %s") % str(e))

        if "settings" not in data:
            return response_error(_("Wrong JSON format. Settings should be set"))

        if not isinstance(data["settings"], dict):
            return response_error(
                _("Wrong JSON format. Settings should be a dictionary")
            )

        errors, applied, skipped = preview_data(data, request.user)

        preview_on_site_allowed = (
            request.user.has_perm("content_settings.can_preview_on_site")
            and PREVIEW_ON_SITE_SHOW
        )
        if request.POST.get("_import"):
            if (
                "content_settings_form_checksum" in request.POST
                and ADMIN_CHECKSUM_CHECK_BEFORE_SAVE
                and request.POST["content_settings_form_checksum"]
                != content_settings.form_checksum
            ):
                return response_error(
                    _(
                        "The content settings have been changed by another user. Please reload the page and try again. (or disable ADMIN_CHECKSUM_CHECK_BEFORE_SAVE in settings)"
                    )
                )

            if request.POST.get("preview_on_site") and not preview_on_site_allowed:
                return response_error(
                    _(
                        "You don't have permissions to apply preview on site. (Technocally it is only possible if the changes are applied just recently)"
                    )
                )

            applied_names = request.POST.getlist("_applied")
            try:
                import_to(
                    data,
                    [row for row in applied if row["name"] in applied_names],
                    request.POST.get("preview_on_site"),
                    request.user,
                )
            except Exception as e:
                return response_error(str(e))

            prepared_names = {}
            for v in applied:
                if v["name"] not in applied_names:
                    continue

                data_setting = data["settings"][v["name"]]
                if "user_defined_type" in data_setting:
                    prepared_names[v["name"]] = (
                        data_setting["value"],
                        data_setting["user_defined_type"],
                        (
                            set(data_setting["tags"].splitlines())
                            if data_setting["tags"]
                            else set()
                        ),
                        data_setting["help"],
                    )
                else:
                    prepared_names[v["name"]] = v["new_value"]

            if request.POST.get("preview_on_site") and not preview_on_site_allowed:
                return response_error(
                    _(
                        "You don't have permissions to apply preview on site. (Technocally it is only possible if the changes are applied just recently)"
                    )
                )

            for name in applied_names:
                if name not in prepared_names:
                    return response_error(_("Setting %s can not be applied") % name)

            try:
                validate_all_with_context(prepared_names)
            except ValidationError as e:
                return response_error(str(e))

            if request.POST.get("preview_on_site"):
                self.message_user(
                    request,
                    _("Settings successfully imported to preview"),
                    messages.SUCCESS,
                )
            else:
                self.message_user(
                    request, _("Settings successfully imported"), messages.SUCCESS
                )

            return HttpResponseRedirect(
                resolve_url("admin:content_settings_contentsetting_changelist")
            )

        context = {
            **core_context,
            "skipped": skipped,
            "applied": applied,
            "errors": errors,
            "raw_json": json.dumps(data, indent=2),
            "preview_on_site_allowed": preview_on_site_allowed,
        }
        return TemplateResponse(
            request, "admin/content_settings/contentsetting/import_json.html", context
        )

    def preview_setting(self, request):
        if request.POST.get("user_defined_type"):
            try:
                cs_type_instance = USER_DEFINED_TYPES_INSTANCE[
                    request.POST["user_defined_type"]
                ]
            except KeyError:
                return JsonResponse(
                    {
                        "error": _("Invalid user_defined_type"),
                        "html": "",
                    }
                )
            else:
                cs_type = cs_type_instance()
        else:
            cs_type = get_type_by_name(request.POST["name"])
        if cs_type is None or cs_type.constant:
            return JsonResponse({"html": "", "error": _("Invalid name")})

        value = request.POST["value"]

        other_values = {
            name[2:]: value
            for name, value in request.POST.items()
            if name.startswith("o_")
        }

        params = {
            name[2:]: value
            for name, value in request.POST.items()
            if name.startswith("p_")
        }

        with content_settings_context(**other_values):
            return JsonResponse(
                cs_type.get_full_admin_preview_value(
                    value, request.POST["name"], user=request.user, **params
                )
            )


class HistoryContentSettingAdmin(admin.ModelAdmin):
    list_display = [
        "created_on",
        "user_defined_type",
        "name",
        "value",
        "version",
        "tags",
        "help",
        "was_changed",
        "user",
    ]
    list_filter = [
        "was_changed",
        "name",
    ]
    search_fields = ["name", "value", "tags", "help"]
    actions = [
        "export_as_json",
    ]

    def export_as_json(self, request, queryset):
        ids = dict(queryset.order_by("-id").values_list("name", "id"))
        return HttpResponseRedirect(
            reverse("admin:content_settings_import_json")
            + f'?history_ids={",".join(map(str, ids.values()))}'
        )

    export_as_json.short_description = _("Export selected history records as JSON")

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, *args):
        return False

    def has_change_permission(self, request, obj=None):
        return False


admin.site.register(ContentSetting, ContentSettingAdmin)
admin.site.register(HistoryContentSetting, HistoryContentSettingAdmin)



"""
the module contains context processors for the django templates.
"""

from .conf import content_settings as _content_settings


def content_settings(request=None):
    """
    context processor for the django templates that provides content_settings object into template as CONTENT_SETTINGS.
    """
    return {"CONTENT_SETTINGS": _content_settings}



# Generated by Django 4.2.9 on 2024-02-24 18:39

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("content_settings", "0002_user_defined"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="contentsetting",
            options={
                "ordering": ("name",),
                "permissions": [("can_preview_on_site", "Can preview on site")],
            },
        ),
        migrations.CreateModel(
            name="UserPreviewHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("created_on", models.DateTimeField(default=django.utils.timezone.now)),
                ("name", models.CharField(max_length=200)),
                ("value", models.TextField(null=True)),
                (
                    "status",
                    models.IntegerField(
                        choices=[
                            (0, "Created"),
                            (10, "Applied"),
                            (20, "Removed"),
                            (30, "Ignored"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserPreview",
            fields=[
                (
                    "id",
                    models.AutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("from_value", models.TextField()),
                ("value", models.TextField()),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("user", "name")},
            },
        ),
    ]



from django import template
from django.utils.html import escape
from django.utils.safestring import SafeString
from django.utils.safestring import mark_safe

from content_settings.conf import content_settings

register = template.Library()


@register.simple_tag
def content_settings_call(name, *args, **kwargs):
    """
    template tag that call callable settings in the template
    """
    safe = kwargs.pop("_safe", False)
    var = getattr(content_settings, name)
    result = var(*args, **kwargs)
    return (
        mark_safe(result) if safe or isinstance(result, SafeString) else escape(result)
    )



"""
The most basic types for the content settings. SimpleString is used as the base for all other types.
"""

from __future__ import annotations

from pprint import pformat
from typing import Optional, Set, Tuple, Union, Any, Callable, Dict
from collections.abc import Iterable
from json import dumps
from inspect import ismethod

from django import forms
from django.contrib.auth.models import User
from django.core.serializers.json import DjangoJSONEncoder
from django.core.exceptions import ValidationError

from content_settings.types.lazy import LazyObject
from content_settings.types import (
    PREVIEW,
    required,
    optional,
    pre,
    BaseSetting,
    TCallableStr,
)
from content_settings.types.mixins import HTMLMixin
from content_settings.defaults.context import update_defaults
from content_settings.utils import call_base_str

from .mixins import EmptyNoneMixin


class SimpleString(BaseSetting):
    """
    A very basic class that returns the string value, the same as a given value.

    Basic attributes (TCallableStr - type for attribute that can be either a string and an import line):

    - `constant: bool = False`: Whether the setting is constant (can not be changed).
    - `cls_field: TCallableStr = "CharField"`: The form field class to use for the setting. For str value class from django.forms.fields is used.
    - `widget: TCallableStr = "LongTextInput"`: The form widget to use for the cls_field. For str value class from django.forms.widgets is used.
    - `widget_attrs: Optional[dict] = None`: Optional attributes for the widget initiation.
    - `fetch_permission: TCallableStr = "none"`: Permission required to fetch the setting in API. For str value function from content_settings.permissions is used
    - `update_permission: TCallableStr = "staff"`: Optional permission required to update  the setting in Django Admin. For str value function from content_settings.permissions is used
    - `view_permission: TCallableStr = "staff"`: Optional permission required to view the setting in Django Admin. For str value function from content_settings.permissions is used
    - `view_history_permission: Optional[TCallableStr]`: Optional permission required to see the hisotry of changes. `None` means the permission is taken from `view_permission`. For str value function from content_settings.permissions is used
    - `help: Optional[str] = ""`: Optional help text for the setting.
    - `value_required: bool = False`: Whether a value is required for the setting.
    - `version: str = ""`: The version of the setting (using for caching).
    - `tags: Optional[Iterable[str]] = None`: Optional tags associated with the setting.
    - `validators: Tuple[TCallableStr] = ()`: Validators to apply to the setting value.
    - `validators_raw: Tuple[TCallableStr] = ()`: Validators to apply to the text value of the setting.
    - `admin_preview_as: PREVIEW = PREVIEW.NONE`: The format to use for the admin preview.

    Advanced attributes - rarely used:

    - `help_format: Optional[str] = "string"`: Optional format string for the help text for the format (align to the type).
    - `suffixes: Optional[Dict[str, TCallableStr]] = None`: Suffixes that can be appended to the setting value.
    - `user_defined_slug: Optional[str] = None`: it contains a slug from db If the setting is defined in DB only (should not be set in content_settings)
    - `overwrite_user_defined: bool = False`: Whether the setting can overwrite a user defined setting.
    - `default: Union[str, required, optional] = ""`: The default value for the setting.
    - `on_change: Tuple[TCallableStr] = ()`: list of functions to call when the setting is changed
    - `on_change_commited: Tuple[TCallableStr] = ()`: list of functions to call when the setting is changed and commited
    - `admin_head_css: Tuple[str] = ()`: list of css urls to include in the admin head
    - `admin_head_js: Tuple[str] = ()`: list of js urls to include in the admin head
    - `admin_head_css_raw: Tuple[str] = ()`: list of css codes to include in the admin head
    - `admin_head_js_raw: Tuple[str] = ()`: list of js codes to include in the admin head
    """

    constant: bool = False
    cls_field: TCallableStr = "CharField"
    widget: TCallableStr = "LongTextInput"
    widget_attrs: Optional[dict] = None
    fetch_permission: TCallableStr = "none"
    update_permission: TCallableStr = "staff"
    view_permission: TCallableStr = "staff"
    view_history_permission: Optional[TCallableStr] = None
    help_format: str = "string"
    help: str = ""
    value_required: bool = False
    version: str = ""
    tags: Optional[Iterable[str]] = None
    validators: Tuple[TCallableStr] = ()
    validators_raw: Tuple[TCallableStr] = ()
    admin_preview_as: PREVIEW = PREVIEW.NONE
    suffixes: Optional[Dict[str, TCallableStr]] = None
    user_defined_slug: Optional[str] = None
    overwrite_user_defined: bool = False
    default: Union[str, required, optional] = ""
    json_encoder: type = DjangoJSONEncoder
    on_change: Tuple[TCallableStr] = ()
    on_change_commited: Tuple[TCallableStr] = ()
    admin_head_css: Tuple[str] = ()
    admin_head_js: Tuple[str] = ()
    admin_head_css_raw: Tuple[str] = ()
    admin_head_js_raw: Tuple[str] = ()

    def __init__(
        self, default: Optional[Union[str, required, optional]] = None, **kwargs
    ):
        """
        The init function accepts initial attributes for the content setting type.
        It can assing any attribute that is defined in the class. (The exception is help_text instead of help)

        All of the changes for type instance can only be done inside of __init__ method. The other methods should not change self object.
        """
        # optional support of help_text instead of help
        assert not (
            "help" in kwargs and "help_text" in kwargs
        ), "You cannot use both help and help_text"
        if "help" not in kwargs and "help_text" in kwargs:
            kwargs["help"] = kwargs.pop("help_text")

        for k in kwargs.keys():
            assert self.can_assign(k), "Attribute {} not found".format(k)

        if default is not None:
            self.default = default
        assert self.default in (required, optional) or isinstance(
            self.default, str
        ), "Default should be str (or required or optional for special cases)"

        updated_kwargs = update_defaults(self, kwargs)
        self.init_assign_kwargs(updated_kwargs)

        assert isinstance(self.version, str), "Version should be str"
        assert (
            self.get_admin_preview_as() in PREVIEW
        ), f"admin_preview_as should be one of {PREVIEW}"

    def init_assign_kwargs(self, kwargs):
        """
        Assign the attributes from the kwargs to the instance.

        Use init__{attribute_name} method if it exists, otherwise use setattr
        """
        for k, v in kwargs.items():
            if not self.can_assign(k):
                raise ValueError("Attribute {} not found".format(k))
            if hasattr(self, "init__{}".format(k)):
                getattr(self, "init__{}".format(k))(v)
            else:
                setattr(self, k, v)

    def init__tags(self, tags: Union[None, str, Iterable[str]]) -> None:
        """
        Assign the tags to the instance from kwargs.
        """
        if not tags:
            return

        if isinstance(tags, str):
            tags = {tags}
        else:
            tags = set(tags)

        self.tags = tags if self.tags is None else self.tags | tags

    def get_admin_head_css(self) -> Tuple[str]:
        """
        Return the list of css urls to include in the admin head.
        """
        return self.admin_head_css

    def get_admin_head_js(self) -> Tuple[str]:
        """
        Return the list of js urls to include in the admin head.
        """
        return self.admin_head_js

    def get_admin_head_css_raw(self) -> Tuple[str]:
        """
        Return the list of css codes to include in the admin head.
        """
        return self.admin_head_css_raw

    def get_admin_head_js_raw(self) -> Tuple[str]:
        """
        Return the list of js codes to include in the admin head.
        """
        return self.admin_head_js_raw

    def can_view(self, user: User) -> bool:
        """
        Return True if the user has permission to view the setting in the django admin panel.

        Use view_permission attribute
        """
        return call_base_str(
            self.view_permission, user, call_base="content_settings.permissions"
        )

    def can_view_history(self, user: User) -> bool:
        """
        Return True if the user has permission to view the setting changing history in the django admin panel.

        Use view_history_permission attribute, but if it is None, use can_view
        """
        return (
            self.can_view(user)
            if self.view_history_permission is None
            else call_base_str(
                self.view_history_permission,
                user,
                call_base="content_settings.permissions",
            )
        )

    def can_update(self, user: User) -> bool:
        """
        Return True if the user has permission to update the setting in the django admin panel.

        Use update_permission attribute
        """
        return call_base_str(
            self.update_permission, user, call_base="content_settings.permissions"
        )

    def can_fetch(self, user: User) -> bool:
        """
        Return True if the user has permission to fetch the setting value using API.

        Use fetch_permission attribute
        """
        return call_base_str(
            self.fetch_permission, user, call_base="content_settings.permissions"
        )

    def get_admin_preview_as(self) -> str:
        """
        Return the format (PREVIEW Enum) to use for the admin preview.

        Use admin_preview_as attribute
        """
        return self.admin_preview_as

    def get_on_change(self) -> Tuple[Callable]:
        """
        Return the list of functions to call when the setting is changed.

        Use on_change attribute
        """
        return self.on_change

    def get_on_change_commited(self) -> Tuple[Callable]:
        """
        Return the list of functions to call when the setting is changed and commited.
        Uses for syncing data or triggering emails.

        Use on_change_commited attribute–
        """
        return self.on_change_commited

    def get_suffixes(self, value: Any) -> Dict[str, Callable]:
        """
        Return the list of suffixes that can be used
        """
        return self.suffixes

    def get_suffixes_names(self, value: Any) -> Tuple[str]:
        """
        Return the tuple of suffixes that can be used
        """
        return tuple(self.get_suffixes(value).keys())

    def can_suffix(self, suffix: Optional[str], value: Any) -> bool:
        """
        Return True if the suffix is valid for the setting.
        """
        return suffix is None or suffix in self.get_suffixes_names(value)

    def can_assign(self, name: str) -> bool:
        """
        Return True if the attribute can be assigned to the instance.
        """
        if name.startswith("_"):
            return False
        if not hasattr(self, name):
            return False
        if ismethod(getattr(self, name)):
            return False
        return True

    def get_help_format(self) -> Iterable[str]:
        """
        Generate help for the specific type.

        The help of the format goes after the help for the specific setting and expalins the format of the setting.
        """
        yield self.help_format

    def get_help(self) -> str:
        """
        Generate help for the specific setting (includes format help)
        """
        help = self.help
        help_format = "".join(self.get_help_format())

        if help and help_format:
            return f"{help}<br><br>{help_format}"

        return help or help_format

    def get_tags(self) -> Set[str]:
        """
        Return the tags associated with the setting.
        """
        tags = self.tags
        if tags is None:
            tags = set()
        return set(tags)

    def get_content_tags(self, name: str, value: str) -> Set[str]:
        """
        Generate tags based on current type and value.

        Uses CONTENT_SETTINGS_TAGS for generating, but overriding the method allows you to change the behavior for the specific type.
        """
        from content_settings.conf import gen_tags

        return gen_tags(name, self, value)

    def get_validators(self) -> Tuple[Callable[[Any], None]]:
        """
        Return the list of validators to apply to the setting python value.
        """
        return tuple(self.validators)

    def get_validators_raw(self) -> Tuple[Callable]:
        """
        Return the list of validators to apply to the setting text value.
        """
        return (
            tuple(self.validators_raw)
            + tuple(self.get_field().default_validators)
            + (("not_empty",) if self.value_required else ())
        )

    def get_field(self) -> forms.Field:
        """
        Generate the form field for the setting. Which will be used in the django admin panel.
        """
        return call_base_str(
            self.cls_field,
            call_base="django.forms.fields",
            widget=self.get_widget(),
            validators=(self.validate_value,),
            required=self.value_required,
        )

    def get_widget(self) -> forms.Widget:
        """
        Generate the form widget for the setting. Which will be used in the django admin panel.
        """
        if isinstance(self.widget, forms.Widget):
            return self.widget
        return call_base_str(
            self.widget,
            call_base="content_settings.widgets",
            attrs=self.widget_attrs,
        )

    def validate_raw_value(self, value: str) -> None:
        """
        Validate the text value of the setting.
        In the validation you only need to make sure the value is possible to be converted into py object.
        """
        if not value and not self.value_required:
            return

        for validator in self.get_validators_raw():
            call_base_str(
                validator, value, call_base="content_settings.types.validators"
            )

    def validate_value(self, value: str) -> Any:
        """
        Full validation of the setting text value.
        """
        self.validate_raw_value(value)
        val = self.to_python(value)
        self.validate(val)

    def validate(self, value: Any):
        """
        Validate py object. Validate consistency of the object with the project.
        """
        for validator in self.get_validators():
            call_base_str(
                validator, value, call_base="content_settings.types.validators"
            )

    def to_python(self, value: str) -> Any:
        """
        Converts text value to python value.
        """
        return self.get_field().to_python(value)

    def json_view_value(self, value: Any, **kwargs) -> Any:
        """
        Converts the setting value to JSON.
        """
        return dumps(value, cls=self.json_encoder)

    def give_python_to_admin(self, value: str, name: str, **kwargs) -> Any:
        """
        Converts the setting text value to setting admin value that will be used for rendering admin preview.

        By default it uses to_python method, but it make sense to override it for some types, for example callable types,
        where you want to show the result of the call in the preview.
        """
        return self.to_python(value)

    def get_admin_preview_html(self, value: Any, name: str, **kwargs) -> Any:
        """
        Generate the admin preview for PREVIEW.HTML format.
        """
        return value

    def get_admin_preview_text(self, value: Any, name: str, **kwargs) -> Any:
        """
        Generate the admin preview for PREVIEW.TEXT format.
        """
        return value

    def get_admin_preview_python(self, value: Any, name: str, **kwargs) -> Any:
        """
        Generate the admin preview for PREVIEW.PYTHON format.
        """
        return value

    def get_admin_preview_value(self, value: str, *args, **kwargs) -> str:
        """
        Generate the admin preview for the setting based on the admin_preview_as attribute (or get_admin_preview_as method).

        Using text value of the setting.
        """
        if self.get_admin_preview_as() == PREVIEW.NONE:
            return ""

        value = self.give_python_to_admin(value, *args, **kwargs)

        return str(self.get_admin_preview_object(value, *args, **kwargs))

    def get_full_admin_preview_value(self, value: str, *args, **kwargs) -> str:
        """
        Generate data for json response for preview
        """
        try:
            self.validate_value(value)
        except Exception as e:
            return {
                "error": str(e),
            }

        return {
            "html": self.get_admin_preview_value(value, *args, **kwargs),
        }

    def get_admin_preview_object(self, value: Any, name: str, **kwargs) -> str:
        """
        Generate the admin preview for the setting based on the admin_preview_as attribute (or get_admin_preview_as method).

        Using admin value of the setting.
        """
        if self.get_admin_preview_as() == PREVIEW.HTML:
            return str(self.get_admin_preview_html(value, name, **kwargs))

        if self.get_admin_preview_as() == PREVIEW.TEXT:
            value = str(self.get_admin_preview_text(value, name, **kwargs))
        else:
            value = pformat(self.get_admin_preview_python(value, name, **kwargs))

        return pre(value)

    def lazy_give(self, l_func: Callable, suffix=None) -> LazyObject:
        """
        Return the LazyObject that will be used for the setting value.

        This value will be returned using lazy prefix in the content_settings.
        """
        return LazyObject(l_func)

    def give(self, value: Any, suffix: Optional[str] = None):
        """
        The will be returned as the content_settings attribute using python value of the setting.

        Suffix can be used.
        """
        if suffix is None:
            return value

        if not self.can_suffix(suffix, value):
            raise KeyError(f"Suffix {suffix} is not allowed")

        return call_base_str(self.get_suffixes(value)[suffix], value)

    def give_python(self, value: str, suffix=None) -> Any:
        return self.give(self.to_python(value), suffix)


class SimpleText(SimpleString):
    """
    Multiline text setting type.
    """

    widget: TCallableStr = "LongTextarea"


class SimpleTextPreview(SimpleText):
    """
    Multiline text setting type with preview. By default SimpleText and SimpleString don't have preview, but for showing preview in EachMixin, we need to have preview for each type.
    """

    admin_preview_as: PREVIEW = PREVIEW.TEXT


class SimpleHTML(HTMLMixin, SimpleText):
    """
    Multiline HTML setting type.
    """

    pass


class URLString(EmptyNoneMixin, SimpleString):
    """
    URL setting type.
    """

    cls_field: TCallableStr = "URLField"
    widget: TCallableStr = "LongURLInput"
    help_format: str = "URL"

    def json_view_value(self, value: Any, **kwargs) -> Any:
        return "null" if value is None else f'"{value}"'


class EmailString(EmptyNoneMixin, SimpleString):
    """
    Email setting type.
    """

    cls_field: TCallableStr = "EmailField"
    widget: TCallableStr = "LongEmailInput"
    help_format: str = "Email"

    def json_view_value(self, value: Any, **kwargs) -> Any:
        return "null" if value is None else f'"{value}"'


class SimpleInt(SimpleString):
    """
    Integer setting type.
    """

    admin_preview_as: PREVIEW = PREVIEW.PYTHON
    cls_field: TCallableStr = "IntegerField"
    widget: TCallableStr = "NumberInput"
    help_format: str = "Any number"

    def json_view_value(self, value: Any, **kwargs) -> Any:
        return str(value)


class SimpleBool(SimpleString):
    """
    Boolean setting type.

    Attributes:
    - yeses (Tuple[str]): Accepted values for True.
    - noes (Tuple[str]): Accepted values for False.
    """

    admin_preview_as: PREVIEW = PREVIEW.PYTHON
    widget: TCallableStr = "TextInput"
    yeses: Tuple[str] = ("yes", "true", "1", "+", "ok")
    noes: Tuple[str] = ("no", "not", "false", "0", "-", "")

    def to_python(self, value) -> bool:
        value = value.lower().strip()
        if value in self.yeses:
            return True
        if value in self.noes:
            return False
        return None

    @property
    def explained_accepted_values(self):
        return [f"'{v}'" if v else "empty" for v in self.yeses + self.noes]

    def validate(self, value):
        if value is None:
            raise ValidationError(
                f"Value cannot be {', '.join(self.explained_accepted_values)} only"
            )

    def get_help_format(self):
        yield "boolean (True/False) value. "
        yield f"Accepted values: {', '.join(self.explained_accepted_values)}"

    def json_view_value(self, value: Any, **kwargs) -> Any:
        return "true" if value else "false"


class SimpleDecimal(SimpleString):
    """
    Decimal setting type.

    Attributes:
    - decimal_json_as_string (bool): set False if you want to return the decimal as a float in the JSON view.
    """

    admin_preview_as: PREVIEW = PREVIEW.PYTHON
    cls_field: TCallableStr = "DecimalField"
    widget: TCallableStr = "TextInput"
    help_format: str = "Decimal number with floating point"
    decimal_json_as_string: bool = True

    def json_view_value(self, value: Any, **kwargs) -> Any:
        return f'"{value}"' if self.decimal_json_as_string else str(value)


class SimplePassword(SimpleString):
    """
    Password setting type. It is not possible to fetch the value using API. In the admin panel, the value is hidden.
    """

    admin_preview_as: PREVIEW = PREVIEW.NONE
    widget: TCallableStr = "PasswordInput"



# Generated by Django 4.2.9 on 2024-01-16 11:35

from django.db import migrations, models
import django.core.validators


class Migration(migrations.Migration):
    dependencies = [
        ("content_settings", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="contentsetting",
            name="help",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="contentsetting",
            name="tags",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="contentsetting",
            name="user_defined_type",
            field=models.CharField(
                db_index=True, default=None, max_length=50, null=True
            ),
        ),
        migrations.AddField(
            model_name="historycontentsetting",
            name="help",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="historycontentsetting",
            name="tags",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="historycontentsetting",
            name="user_defined_type",
            field=models.CharField(default=None, max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name="contentsetting",
            name="name",
            field=models.CharField(
                max_length=200,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        message="Value must be uppercase, can include numbers and underscores, and cannot start with a number.",
                        regex="^[A-Z][A-Z0-9_]*$",
                    )
                ],
            ),
        ),
    ]



"""
the module is used for connecting signals to the content settings.
"""

from django.db.models.signals import post_save, post_migrate, post_delete, pre_save
from django.core.signals import request_started
from django.db.backends.signals import connection_created
from django.apps import apps
from django.db import transaction

from .settings import (
    USER_DEFINED_TYPES,
    UPDATE_DB_VALUES_BY_MIGRATE,
    CHECK_UPDATE_CELERY,
    CHECK_UPDATE_HUEY,
)

from django.dispatch import receiver

from .caching import (
    check_update,
    recalc_checksums,
    validate_default_values,
)
from .conf import set_initial_values_for_db, get_type_by_name, get_str_tags
from .models import ContentSetting, HistoryContentSetting
from .utils import call_base_str


@receiver(post_delete, sender=ContentSetting)
@receiver(post_save, sender=ContentSetting)
def do_update_stored_checksum(*args, **kwargs):
    """
    update the stored checksum of the settings.
    """
    connection = transaction.get_connection()
    if connection.in_atomic_block:
        transaction.on_commit(recalc_checksums)
    else:
        recalc_checksums()


@receiver(post_save, sender=ContentSetting)
def trigger_on_change(sender, instance, created, **kwargs):
    """
    Trigger on_change and on_change_commited for the content setting
    """
    if created:
        return

    cs_type = get_type_by_name(instance.name)
    if cs_type is None:
        return

    for on_change in cs_type.get_on_change():
        call_base_str(on_change, instance)

    on_changes = cs_type.get_on_change_commited()
    if not on_changes:
        return

    connection = transaction.get_connection()
    if connection.in_atomic_block:
        transaction.on_commit(lambda: [call_base_str(f, instance) for f in on_changes])
    else:
        for on_change in on_changes:
            call_base_str(on_change, instance)


@receiver(post_save, sender=ContentSetting)
def create_history_settings(sender, instance, created, **kwargs):
    HistoryContentSetting.objects.create(
        name=instance.name,
        value=instance.value,
        version=instance.version,
        tags=instance.tags,
        help=instance.help,
        user_defined_type=instance.user_defined_type,
        was_changed=not created,
    )


@receiver(pre_save, sender=ContentSetting)
def update_value_tags(sender, instance, **kwargs):
    if instance.user_defined_type:
        return

    cs_type = get_type_by_name(instance.name)
    if cs_type is None:
        return

    instance.tags = get_str_tags(instance.name, cs_type, instance.value)


if USER_DEFINED_TYPES:

    @receiver(pre_save, sender=ContentSetting)
    def check_user_defined_type_version(sender, instance, **kwargs):
        from .conf import USER_DEFINED_TYPES_INITIAL

        if instance.user_defined_type not in USER_DEFINED_TYPES_INITIAL:
            return
        instance.version = USER_DEFINED_TYPES_INITIAL[
            instance.user_defined_type
        ].version
        if instance.tags:
            instance.tags = instance.tags.replace("\r\n", "\n")


@receiver(post_delete, sender=ContentSetting)
def create_history_settings_delete(sender, instance, **kwargs):
    HistoryContentSetting.objects.create(
        name=instance.name,
        value=instance.value,
        version=instance.version,
        tags=instance.tags,
        help=instance.help,
        was_changed=None,
    )


@receiver(connection_created)
def validate_default_values_for_connection(*args, **kwargs):
    validate_default_values()


@receiver(request_started)
def check_update_for_request(*args, **kwargs):
    check_update()


if UPDATE_DB_VALUES_BY_MIGRATE:

    @receiver(post_migrate, sender=apps.get_app_config("content_settings"))
    def update_stored_values(*args, **kwargs):
        try:
            with transaction.atomic():
                log = set_initial_values_for_db(apply=True)
        except Exception as e:
            log = [("Error", str(e))]

        if not log:
            return

        for l in log:
            print(" ".join(l))


# INTEGRATIONS

if CHECK_UPDATE_CELERY:
    try:
        from celery.signals import task_prerun
    except ImportError:
        pass
    else:

        @task_prerun.connect
        def check_update_for_celery(*args, **kwargs):
            check_update()


if CHECK_UPDATE_HUEY:
    try:
        from huey.contrib.djhuey import pre_execute
    except ImportError:
        pass
    else:

        @pre_execute()
        def check_update_for_huey(*args, **kwargs):
            check_update()



import json
from django.core.management.base import BaseCommand
from content_settings.models import ContentSetting
from content_settings.export import export_to_format


class Command(BaseCommand):
    help = "Export selected content settings (by name) or all content settings into JSON format."

    def add_arguments(self, parser):
        parser.add_argument(
            "--names",
            nargs="+",
            type=str,
            help="Names of the content settings to export. If not provided, all content settings will be exported.",
        )

    def handle(self, *args, **options):
        names = options["names"]
        if names:
            content_settings = ContentSetting.objects.filter(name__in=names)
        else:
            content_settings = ContentSetting.objects.all()

        json_data = export_to_format(content_settings)

        self.stdout.write(json_data)



"""
One of the most complicated module contains callable types. The Python object for those types usually needs to be called.

The complexity of the module also illustrates the flexibility of types.

The module is called a "template" because the setting's raw value is a template that will be used to generate a real value.

See `CallToPythonMixin`
"""

from itertools import zip_longest
from typing import Any, Tuple, Optional, Callable, Iterable, Union, Dict
from enum import Enum, auto

from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe
from django.db.models.query import QuerySet
from django.utils.translation import gettext as _

from content_settings.utils import obj_base_str
from .basic import SimpleText
from .mixins import CallToPythonMixin, GiveCallMixin, HTMLMixin
from .validators import call_validator, gen_signle_arg_call_validator
from . import PREVIEW, required


BUILTINS_SAFE = {}
BUILTINS_ALLOW_IMPORT = {}


def gen_builtins():
    if isinstance(BUILTINS_SAFE, dict):
        yield from __builtins__.items()
    else:
        for name in dir(__builtins__):
            yield name, getattr(__builtins__, name)


for name, value in gen_builtins():
    if name in ["memoryview", "open", "input"]:
        continue
    BUILTINS_ALLOW_IMPORT[name] = value

    if name == "__import__":
        continue
    BUILTINS_SAFE[name] = value


class STATIC_INCLUDES(Enum):
    CONTENT_SETTINGS = auto()
    SETTINGS = auto()


class StaticDataMixin:
    """
    Adds static data to the context, such as SETTINGS or/and CONTENT_SETTINGS.

    Attributes:

    - `template_static_includes` - tuple of `STATIC_INCLUDES` that should be included in the context. Default: `(STATIC_INCLUDES.CONTENT_SETTINGS, STATIC_INCLUDES.SETTINGS)`. If `STATIC_INCLUDES.SETTINGS` is included, `django.conf.settings` will be added to the context. If `STATIC_INCLUDES.CONTENT_SETTINGS` is included, `content_settings.conf.content_settings` will be added to the context.
    - `template_static_data` - static data that should be added to the context (on top of what will be added by `template_static_includes`). It can be a dictionary or a callable that returns a dictionary. Default: `None`.
    """

    template_static_includes: Tuple[STATIC_INCLUDES] = (
        STATIC_INCLUDES.CONTENT_SETTINGS,
        STATIC_INCLUDES.SETTINGS,
    )
    template_static_data: Optional[Union[Dict, Callable]] = None

    def get_template_static_data(self):
        if not self.template_static_data:
            return {}
        if callable(self.template_static_data):
            return self.template_static_data()
        return self.template_static_data

    def get_template_full_static_data(self):
        ret = {}

        if STATIC_INCLUDES.CONTENT_SETTINGS in self.template_static_includes:
            from content_settings.conf import content_settings

            ret["CONTENT_SETTINGS"] = content_settings

        if STATIC_INCLUDES.SETTINGS in self.template_static_includes:
            from django.conf import settings

            ret["SETTINGS"] = settings

        return {
            **ret,
            **self.get_template_static_data(),
        }


class SimpleFunc(CallToPythonMixin, SimpleText):
    admin_preview_as: PREVIEW = PREVIEW.PYTHON


class SimpleCallTemplate(CallToPythonMixin, StaticDataMixin, SimpleText):
    """
    Base class for templates that can be called.
    """

    admin_preview_as: PREVIEW = PREVIEW.TEXT
    template_args_default = None

    def get_template_args_default(self):
        if not self.template_args_default:
            return {}
        return self.template_args_default

    def get_validators(self):
        validators = super().get_validators()

        if validators and any(isinstance(v, call_validator) for v in validators):
            return validators

        has_required_args = any(
            v == required for v in self.get_template_args_default().values()
        )
        if has_required_args:
            return validators

        return (call_validator(), *validators)

    def get_help_format(self):
        yield _("Available objects:")
        yield "<ul>"

        for name, default in self.get_template_args_default().items():
            yield f"<li>{name}"
            if default == required:
                yield " - "
                yield _("required")
                yield "</li>"
            else:
                yield " - "
                yield repr(default)
                yield "</li>"

        for name in self.get_template_full_static_data().keys():
            yield f"<li>{name}</li>"

        yield "</ul>"

    def prepate_input_to_dict(self, *args, **kwargs):
        """
        prepares an inpit dictuonary for the call based on the given arguments and kwargs
        """
        template_args_default = self.get_template_args_default()
        no_value = {}
        kwargs = {
            **{
                default[0]: (default[1] if arg_value == no_value else arg_value)
                for arg_value, default in zip_longest(
                    args, template_args_default.items(), fillvalue=no_value
                )
                if default != no_value
            },
            **kwargs,
        }

        for name, value in kwargs.items():
            if value == required:
                raise ValidationError(
                    _("Missing required argument %(name)s") % {"name": name}
                )

        return kwargs


class DjangoTemplate(SimpleCallTemplate):
    """
    The setting of that type generates value based on the Django Template in the raw value.
    """

    admin_preview_as: PREVIEW = PREVIEW.TEXT

    def get_help_format(self):
        yield _(
            "Simple <a href='https://docs.djangoproject.com/en/3.2/topics/templates/' target='_blank'>Django Template</a>."
        )
        yield " "
        yield from super().get_help_format()

    def prepare_python_call(self, value):
        from django.template import Template

        return {"template": Template(value.strip())}

    def python_call(self, *args, **kwargs):
        from django.template import Context

        template = kwargs.pop("template")
        return (
            template.render(
                Context(
                    {
                        **self.get_template_full_static_data(),
                        **self.prepate_input_to_dict(*args, **kwargs),
                    }
                )
            )
            + ""  # to avoid returning SafeText
        )


class DjangoTemplateHTML(HTMLMixin, DjangoTemplate):
    """
    Same as `DjangoTemplate` but the rendered value is marked as safe.
    """

    pass


class DjangoTemplateNoArgs(GiveCallMixin, DjangoTemplate):
    """
    Same as `DjangoTemplate` but the setting value is not callablle, but already rendered value.
    """

    pass


class DjangoTemplateNoArgsHTML(HTMLMixin, DjangoTemplateNoArgs):
    """
    Same as `DjangoTemplateNoArgs` but the rendered value is marked as safe.
    """

    pass


class DjangoModelTemplateMixin:
    """
    Mixing that uses one argument for the template from the model queryset.

    Attributes:

    - `template_model_queryset` - QuerySet or a callable that returns a Model Object. For QuerySet, the first object will be used. For callable, the object returned by the callable will be used. The generated object will be used as validator and for preview.
    - `template_object_name` - name of the object in the template. Default: "object".
    """

    template_model_queryset: Optional[Union[QuerySet, Callable]] = None
    template_object_name: str = "object"

    def get_template_args_default(self):
        return {
            self.template_object_name: required,
            **super().get_template_args_default(),
        }

    def get_first_call_validator(self):
        """
        generates the first validator based of template_model_queryset, which will be used for validation and for preview.
        """
        if self.template_model_queryset is not None:
            if isinstance(self.template_model_queryset, QuerySet):
                first = self.template_model_queryset.first()
                if first is not None:
                    return call_validator(first)
            elif callable(self.template_model_queryset):
                return gen_signle_arg_call_validator(self.template_model_queryset)
            else:
                raise ValueError(
                    _("template_model_queryset must be a QuerySet or a callable")
                )

    def get_validators(self):
        validators = super().get_validators()
        first_validator = self.get_first_call_validator()

        if first_validator is not None:
            validators = (first_validator, *validators)

        return validators


class DjangoModelTemplate(DjangoModelTemplateMixin, DjangoTemplate):
    """
    Django Template that uses one argument as a model object.
    """

    pass


class DjangoModelTemplateHTML(DjangoModelTemplate):
    """
    Same as `DjangoModelTemplate` but the rendered value is marked as safe.
    """

    admin_preview_as: PREVIEW = PREVIEW.HTML

    def give(self, value, suffix=None):
        render_func = super().give(value, suffix)
        return lambda *a, **k: mark_safe(render_func(*a, **k))


class SimpleEval(SimpleCallTemplate):
    """
    Template that evaluates the Python code (using `eval` function).
    """

    admin_preview_as: PREVIEW = PREVIEW.PYTHON
    template_bultins: Optional[Union[str, Dict]] = "BUILTINS_SAFE"

    def get_help_format(self):
        yield _("Python code that returns a value.")
        yield " "
        yield from super().get_help_format()

    def prepare_python_call(self, value):
        return {"template": compile(value, "<string>", "eval")}

    def python_call(self, *args, **kwargs):
        template = kwargs.pop("template")
        globs = {
            **self.get_template_full_static_data(),
            **self.prepate_input_to_dict(*args, **kwargs),
        }

        if self.template_bultins:
            globs["__builtins__"] = obj_base_str(
                self.template_bultins, "content_settings.types.template"
            )

        return eval(template, globs)


class DjangoModelEval(DjangoModelTemplateMixin, SimpleEval):
    """
    Same as `SimpleEval` but uses one value as a model object.
    """

    pass


class SimpleEvalNoArgs(GiveCallMixin, SimpleEval):
    """
    Same as `SimpleEval` but the setting value is not callable, but already evaluated.
    """

    pass


class SystemExec:
    """
    Mixin class that brings exec functionality into type

    Attributes:

    - `template_return` - dictates what will be returned as a setting value.
        - If `None`, the whole context will be returned. The default value.
        - if a string, the string will be used as a key from the context and that value will be returned.
        - If a dictionary, only the keys from the dictionary will be returned. The values will be used as defaults.
        - If a callable, the callable will be called and the return value will be used as a dictionary.
        - If an iterable, the iterable will be used as keys for the return dictionary. The values will be `None` by default.
    - `template_bultins` - allows to limit the builtin functions. By default it is `"BUILTINS_SAFE"`, which allows all functions except `memoryview`, `open`, `input` and `import`. If you want to include import use `"BUILTINS_ALLOW_IMPORT"`, if you don't want to use any limitation - just assign `None` to the attribute.
    - `template_raise_return_error` - raises an error if `template_return` doesn't match the context. Default: `False`.
    """

    template_return: Optional[Union[Iterable, Callable, Dict, str]] = None
    template_bultins: Optional[Union[str, Dict]] = "BUILTINS_SAFE"
    template_raise_return_error: bool = False

    def get_template_return(self):
        return self.template_return

    def get_template_raise_return_error(self):
        return self.template_raise_return_error

    def get_template_bultins(self):
        return obj_base_str(self.template_bultins, "content_settings.types.template")

    def get_help_format(self):
        yield from super().get_help_format()

        template_return = self.get_template_return()

        if template_return is None or callable(template_return):
            yield _("Return Dict is not specified")
        elif isinstance(template_return, str):
            yield _("Return value from the variable: %(variable)s") % {
                "variable": template_return
            }
        elif isinstance(template_return, dict):
            yield _("Return Dict: ")
            yield "<ul>"
            for name, value in template_return.items():
                yield f"<li>{name} - default: {value}</li>"
            yield "</ul>"
        elif isinstance(template_return, Iterable):
            yield _("Return Dict: ")
            yield "<ul>"
            for name in template_return:
                yield f"<li>{name}</li>"
            yield "</ul>"
        else:
            raise ValueError(_("Invalid template_return"))

    def system_exec(self, code, globs):
        globs = {**globs}
        if self.template_bultins:
            globs["__builtins__"] = self.get_template_bultins()

        exec(code, globs)
        template_return = self.get_template_return()

        if template_return is None:
            return globs
        elif callable(template_return):
            return template_return(globs)
        elif isinstance(template_return, str):
            if self.get_template_raise_return_error() and template_return not in globs:
                raise ValidationError(
                    _("Variable %(variable)s is required")
                    % {"variable": template_return}
                )
            return globs.get(template_return)
        elif isinstance(template_return, dict):
            return {k: globs.get(k, v) for k, v in template_return.items()}
        elif isinstance(template_return, Iterable):
            if self.get_template_raise_return_error() and (
                diff_keys := set(template_return).difference(set(globs.keys()))
            ):
                raise ValidationError(
                    _("Variables in the context are required: %(variables)s")
                    % {"variables": ", ".join(diff_keys)}
                )
            return {k: globs.get(k, None) for k in template_return}
        else:
            raise ValidationError(_("Invalid template_return"))


class SimpleExec(SystemExec, SimpleCallTemplate):
    """
    Template that executes the Python code (using `exec` function).
    """

    admin_preview_as: PREVIEW = PREVIEW.PYTHON

    def get_help_format(self):
        yield _("Python code that execute and returns generated variables.")
        yield " "
        yield from super().get_help_format()

    def prepare_python_call(self, value):
        return {"template": compile(value, "<string>", "exec")}

    def python_call(self, *args, **kwargs):
        template = kwargs.pop("template")
        globs = {
            **self.get_template_full_static_data(),
            **self.prepate_input_to_dict(*args, **kwargs),
        }

        return self.system_exec(template, globs)


class SimpleExecNoArgs(GiveCallMixin, SimpleExec):
    """
    Same as `SimpleExec` but the setting value is not callable, but already executed.
    """

    pass


class DjangoModelExec(DjangoModelTemplateMixin, SimpleExec):
    """
    Same as `SimpleExec` but uses one value as a model object.
    """

    pass


class SimpleExecNoCompile(SystemExec, StaticDataMixin, SimpleText):
    """
    It is not a Template, probably closed to markdown module, as it simply takes the text value and convert it into py object, which is not a compiled code but the result of execution of the code.

    But it is in the template module as it is very similar to `SimpleExecNoArgs`.

    It is super easy to shot in the foot with this class so be very cautious.
    """

    admin_preview_as: PREVIEW = PREVIEW.PYTHON

    def to_python(self, value: str) -> Any:
        value = super().to_python(value)
        return self.system_exec(value, self.get_template_full_static_data())



"""
A list of functions that are used as values for validators attribute of a type.
"""

from django.core.exceptions import ValidationError
from pprint import pformat


def not_empty(value: str):
    if not value:
        raise ValidationError("Value cannot be empty")


class PreviewValidator:
    """
    return instance of the class from the validator to show the result of validation
    """

    def __init__(self, preview_input: str, preview_output: str):
        self.preview_input = preview_input
        self.preview_output = preview_output


class PreviewValidationError(ValidationError):
    """
    use class instead of ValidatorError to show the input arguments that cause the ValidationError
    """

    def __init__(self, preview_input: str, *args, **kwargs) -> None:
        self.preview_input = preview_input
        super().__init__(*args, **kwargs)


class call_validator:
    """
    create a valiator that calls the function with the given args and kwargs.
    """

    has_call_representation = True

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, func):
        try:
            return func(*self.args, **self.kwargs)
        except Exception as e:
            raise ValidationError(str(e))

    def __str__(self) -> str:
        ret = ""
        if self.args:
            ret += ", ".join([pformat(arg) for arg in self.args])

        if self.kwargs:
            if ret:
                ret += ", "
            ret += ", ".join(
                [f"{key}={pformat(value)}" for key, value in self.kwargs.items()]
            )

        return ret

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {str(self)}>"


class result_validator(call_validator):
    """
    Not only call the function, but also validate the result of the call.

    takes two new arguments:

    * function that validates the result of the call
    * error message that will be shown if the function returns False
    """

    def __init__(self, result_validator, error_message, *args, **kwargs):
        self.result_validator = result_validator
        self.error_message = error_message
        super().__init__(*args, **kwargs)

    def __call__(self, func):
        ret = super().__call__(func)
        if not self.result_validator(ret):
            raise ValidationError(self.error_message)
        return ret


class gen_call_validator(call_validator):
    """
    Same as call_validator, but the args and kwargs are generated by a given function.

    Create a valiator that calls the function that generates the args and kwargs, that will be used for the call.

    The function will be regenerated every time when the validator is called.

    The reason of having one is when you are not able to get possible args at the time of the creation of the validator.
    """

    def __init__(self, gen_args_kwargs_func):
        self.gen_args_kwargs_func = gen_args_kwargs_func
        self._args_kwargs = None

    def gen_args_kwargs(self):
        return self.gen_args_kwargs_func()

    @property
    def args_kwargs(self):
        if self._args_kwargs is None:
            self._args_kwargs = self.gen_args_kwargs()

            assert (
                len(self._args_kwargs) == 2
            ), "gen_args_kwargs_func must return a tuple of args and kwargs"
            assert isinstance(
                self._args_kwargs[0], (list, tuple)
            ), "args must be a list or tuple"
            assert isinstance(self._args_kwargs[1], dict), "kwargs must be a dict"
        return self._args_kwargs

    @property
    def args(self):
        return self.args_kwargs[0]

    @property
    def kwargs(self):
        return self.args_kwargs[1]

    def __call__(self, func):
        try:
            return super().__call__(func)
        finally:
            self._args_kwargs = None


class gen_args_call_validator(gen_call_validator):
    """
    Same as gen_call_validator, but only generates the list for args.
    """

    def gen_args_kwargs(self):
        return self.gen_args_kwargs_func(), {}


class gen_signle_arg_call_validator(gen_call_validator):
    """
    Same as gen_call_validator, but only generates one arg.
    """

    def gen_args_kwargs(self):
        return [self.gen_args_kwargs_func()], {}


class gen_kwargs_call_validator(gen_call_validator):
    """
    Same as gen_call_validator, but only generates the dict for kwargs.
    """

    def gen_args_kwargs(self):
        return [], self.gen_args_kwargs_func()
